<?php

namespace Drupal\stripe_subscription\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Url;
use Drupal\Core\Utility\Token;
use Drupal\stripe_api\StripeApiService;
use Drupal\stripe_subscription\Event\StripeCreateSubscribeSessionEvent;
use Drupal\subscription_manager\SubscriptionManagerService;
use Drupal\token\TokenEntityMapperInterface;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Class UserSubscriptionsController.
 *
 * @package Drupal\stripe_subscription\Controller
 */
class UserSubscriptionsController extends ControllerBase {

  /**
   * @var \Drupal\subscription_manager\SubscriptionManagerService
   */
  protected $subscriptionManager;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  private $logger;

  /**
   * @var \Drupal\stripe_api\StripeApiService
   */
  private $stripeApi;

  private Token $tokenService;

  public function __construct(
    SubscriptionManagerService $subscription_manager,
    LoggerChannelInterface $logger,
    AccountProxyInterface $current_user,
    StripeApiService $stripe_api_service,
    Token $token_service,
  ) {
    $this->subscriptionManager = $subscription_manager;
    $this->logger = $logger;
    $this->currentUser = $current_user;
    $this->stripeApi = $stripe_api_service;
    $this->tokenService = $token_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('subscription_manager'),
      $container->get('logger.channel.stripe_subscription'),
      $container->get('current_user'),
      $container->get('stripe_api.stripe_api'),
      $container->get('token')
    );
  }

  /**
   * @return string
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function subscribeTitle(): string {
    if ($this->subscriptionManager->userHasLocalSubscriptionToAnyPlan($this->currentUser())) {
      return 'Upgrade';
    }
    return 'Subscribe';
  }

  /**
   * @return array
   *   Return
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException|\Drupal\Component\Plugin\Exception\PluginException|\Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function subscribe() {
    if ($this->config('subscription_manager.settings')->get('custom_subscribe_url')) {
      $url = $this->config('subscription_manager.settings')->get('custom_subscribe_url');
      $data = ['user' => $this->currentUser()];
      $options = ['clear' => TRUE];

      // Replace tokens in the URL.
      $url = $this->tokenService->replace($url, $data, $options, new BubbleableMetadata());

      // Encode the URL.
      $encoded_url = Url::fromUri($url)->toString();
      $response = new TrustedRedirectResponse($encoded_url);
      $response->getCacheableMetadata()->addCacheContexts(['user']);

      return $response;
    }

    return $this->renderPricingTableBlock();
    // Return $this->renderLocalPlans();
  }

  /**
   * @param \Symfony\Component\HttpFoundation\Request $request
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   A Response object.
   * @throws \Exception
   */
  public function createSubscribeSession(Request $request): Response {
    if ($this->currentUser->isAnonymous()) {
      $success_url = Url::fromRoute('stripe_subscription.login', [], [
        'absolute' => TRUE,
      ])->toString();
      // We cannot add this to fromRoute() because it will be encoded.
      $success_url = $success_url . '?session_id={CHECKOUT_SESSION_ID}';
      $cancel_url = Url::fromUri($request->getUri(),
        ['absolute' => TRUE, 'query' => ['checkout' => 'failure', 'connector' => 'stripe']])->toString();
    }
    elseif ($request->get('return_url')) {
      $success_url = Url::fromUri(
        'internal:/' . $request->get('return_url'),
        [
          'absolute' => TRUE,
          'query' => [
            'checkout' => 'success',
            'price_id' => $request->get('price_id'),
          ],
        ])->toString();
      $cancel_url = Url::fromUri(
        'internal:/' . $request->get('return_url'),
        [
          'absolute' => TRUE,
          'query' => [
            'checkout' => 'failure',
            'price_id' => $request->get('price_id'),
          ],
        ])->toString();
    }
    else {
      $success_url = Url::fromRoute('<front>', [],
        ['absolute' => TRUE, 'query' => ['checkout' => 'success', 'connector' => 'stripe']])->toString();
      $cancel_url = Url::fromRoute('<front>', [],
        ['absolute' => TRUE, 'query' => ['checkout' => 'failure', 'connector' => 'stripe']])->toString();
    }

    $params = [
      'payment_method_types' => ['card'],
      'line_items' => [
        [
          'quantity' => 1,
          // This must correspond to an existing price id in the Stripe backend.
          'price' => $request->get('price_id'),
        ],
      ],
      'mode' => 'subscription',
      'subscription_data' => [
        'trial_from_plan' => TRUE,
      ],
      'success_url' => $success_url,
      'cancel_url' => $cancel_url,
    ];
    if ($this->currentUser->isAuthenticated()) {
      $params['metadata'] = [
        'module' => 'stripe_subscription',
        'uid' => $this->currentUser()->id(),
      ];
      if ($customer_id = $this->subscriptionManager->getLocalUserCustomerIdFromSubscription(User::load($this->currentUser()->id()), 'stripe_connector')) {
        $params['customer'] = $customer_id;
      }
      else {
        $params['customer_email'] = $this->currentUser()->getEmail();
      }
    }
    if ($request->get('coupon')) {
      // @todo Check that coupon is valid, sending an invalid one will retun
      // an error.
      $params['discounts'][]['coupon'] = $request->get('coupon');
    }
    elseif ($request->get('promo')) {
      $params['discounts'][]['promotion_code'] = $request->get('promo');
    }
    else {
      $params['allow_promotion_codes'] = TRUE;
    }

    $event = new StripeCreateSubscribeSessionEvent($this->currentUser(), $params);
    $event_dispatcher = \Drupal::service('event_dispatcher');
    $event_dispatcher->dispatch($event, StripeCreateSubscribeSessionEvent::EVENT_NAME);

    try {
      $stripe_client = $this->stripeApi->getStripeClient();
      $session = $stripe_client->checkout->sessions->create($params);
    }
    catch (\Exception $exception) {
      $this->logger->error($exception->getMessage());
      return new Response($exception->getMessage(), Response::HTTP_BAD_REQUEST);
    }

    $response_content = [
      'session_id' => $session->id,
      'public_key' => \Drupal::service('stripe_api.stripe_api')->getPubKey(),
    ];

    return new Response(json_encode($response_content), Response::HTTP_ACCEPTED);
  }

  /**
   * @param \Symfony\Component\HttpFoundation\Request $request
   *
   * @return \Symfony\Component\HttpFoundation\Response
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loginFromPurchase(Request $request): Response {
    if (!$request->get('session_id')) {
      $message = 'You must pass a session_id parameter to this route in your request';
      $this->logger->error($message);
      return new Response($message, Response::HTTP_BAD_REQUEST);
    }

    try {
      $stripe_client = $this->stripeApi->getStripeClient();
      $session = $stripe_client->checkout->sessions->retrieve($request->get('session_id'));
      if ($session->status != 'complete') {
        throw new \Exception("Session status is not complete");
      }
      $subscription = $stripe_client->subscriptions->retrieve($session->subscription);
      $plan_id = $subscription->plan->id;
      if (!$plan_id) {
        throw new \Exception("Unable to determine plan id.");
      }
    }
    catch (\Exception $e) {
      $this->logger->error($e->getMessage());
      return new Response($e->getMessage(), Response::HTTP_BAD_REQUEST);
    }
    $customer = $stripe_client->customers->retrieve($session->customer);
    $email = $customer->email;

    if ($this->currentUser->isAuthenticated()) {
      if ($this->currentUser()->getEmail() != $email) {
        $this->messenger()->addError($this->t('You just subscribed with email address @email_purchased, ' .
          'but you\'re logged in with email address @email_logged_in. Please log out and log back' .
          ' in with @email_purchased to access your subscription.',
          ['@email_purchased' => $email, '@email_logged_in' => $this->currentUser()->getEmail()]
        ));
        $this->logger->error('User subscribed with email address @email_purchased, ' .
          'but was already logged in with email address @email_logged_in',
          ['@email_purchased' => $email, '@email_logged_in' => $this->currentUser()->getEmail()]
        );
      }
      else {
        $this->messenger()->addStatus($this->t('Thank you for your purchase! You\'re now subscribed!'));
      }
      return new RedirectResponse(Url::fromRoute('<front>')->toString());
    }

    // We need to make sure that this is the user that just made the purchase, and not someone who just knows the
    // hosted page id.
    $user = user_load_by_mail($email);
    // If this user already exists, send them to the login page. We want this endpoint to be used only for new users.
    // No logged in user should ever actually be redirected to this page anyway.
    // Without this protection, this can be used as a backdoor to access the site using only the hosted page id.
    if ($user) {
      \Drupal::messenger()->addStatus($this->t('Thank you for your purchase. You already have an account for @email. Please login to access your account.', [
        '@email' => $email,
      ]));
      return new RedirectResponse(Url::fromRoute('user.login')->toString());
    }

    // Create user.
    try {
      // @see email_registration_user_insert().
      $new_name = preg_replace('/@.*$/', '', $email);
      $new_name = email_registration_cleanup_username($new_name);
      $new_name = email_registration_unique_username($new_name, 0);

      $fields = [
        'name' => $new_name,
        'mail' => $email,
        'init' => $email,
        'pass' => \Drupal::service('password_generator')->generate(32),
        'status' => 1,
        'field_full_name' => $customer->name,
      ];

      /** @var \Drupal\user\Entity\User $new_user */
      $entity_type_manager = \Drupal::service('entity_type.manager');
      $new_user = $entity_type_manager
        ->getStorage('user')
        ->create($fields);
      $new_user->save();
      $user = $new_user;

      /** @var \Drupal\user\UserDataInterface $userData */
      $user_data = \Drupal::service('user.data');
      // This is used by chargebee_subscription_form_user_form_alter().
      $user_data->set('stripe_subscription', $user->id(), 'has_random_password', TRUE);
    }
    catch (\Exception $e) {
      // If this failed, then the user has purchased a subscription and has no account. That's bad.
      $this->logger->error('User purchased subscription but account creation failed! Email: @email, ' . $e->getMessage(), ['@email' => $email]);
      return new Response($e->getMessage(), Response::HTTP_BAD_REQUEST);
    }

    // Login.
    if ($this->loginUser($user)) {
      $this->messenger()->addStatus($this->t("Thanks for subscribing! Please <u>set your password below</u> and, if you want to, fill out your community profile."));
      return new RedirectResponse(Url::fromRoute('entity.user.edit_form', ['user' => $user->id()])->toString());
    }

    $this->messenger()->addError("Thanks for subscribing! There was an error when attempting to create your account and log you in.");
    $this->logger->error('User purchased subscription and account was created, but user could not be logged in. Email: @email.', ['@email' => $email]);
    return new RedirectResponse(Url::fromRoute('<front>')->toString());
  }

  /**
   * Logs the user in.
   *
   * @param \Drupal\user\Entity\UserInterface $drupal_user
   *   User object.
   *
   * @return bool
   *   True if login was successful
   *   False if the login was blocked
   */
  public function loginUser(UserInterface $drupal_user): bool {
    // Check that the account is active and log the user in.
    if ($drupal_user->isActive()) {
      $this->userLoginFinalize($drupal_user);

      // Dispatches SocialAuthEvents::USER_LOGIN event.
      // $event = new UserEvent($drupal_user, $this->getPluginId());
      // $this->eventDispatcher->dispatch($event, SocialAuthEvents::USER_LOGIN);.
      return TRUE;
    }

    $this->logger->error('Login for user @user prevented. Account is blocked.', ['@user' => $drupal_user->getAccountName()]);

    return FALSE;
  }

  /**
   * Wrapper for user_login_finalize.
   *
   * We need to wrap the legacy procedural Drupal API functions so that we are
   * not using them directly in our own methods. This way we can unit test our
   * own methods.
   *
   * @param \Drupal\user\Entity\UserInterface $account
   *   The Drupal user.
   *
   * @see user_password
   */
  protected function userLoginFinalize(UserInterface $account) {
    user_login_finalize($account);
  }

  /**
   * @return array
   */
  public function renderPricingTableBlock(): array {
    $plugin_block_manager = \Drupal::service('plugin.manager.block');
    $plugin_block = $plugin_block_manager->createInstance('stripe_pricing_table_block');
    $access_result = $plugin_block->access($this->currentUser->getAccount());
    if ((is_bool($access_result) && !$access_result) || (is_object($access_result) && $access_result->isForbidden())) {
      $rendered = [];
    }
    else {
      $rendered = $plugin_block->build($this->config('stripe_subscription.settings')
        ->get('pricing_table_id'));
    }
    return $rendered;
  }

  /**
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  public function renderLocalPlans(): array {
    $local_plans = $this->subscriptionManager->loadLocalPlanMultiple(
      ['status' => 1, 'connector_plugin_id' => 'stripe_connector']
    );
    $user = User::load($this->currentUser->id());
    $build = [
      '#theme' => 'stripe_subscribe_plans',
      '#plans' => [],
    ];
    foreach ($local_plans as $local_plan) {
      $element = [
        '#theme' => 'stripe_subscribe',
        '#plan' => [
          'price_id' => $local_plan->getData()->first()->getValue()['id'],
          'name' => $local_plan->getName(),
          'interval' => $local_plan->getData()->first()->getValue()['interval'],
          'nickname' => $local_plan->getData()->first()->getValue()['nickname'],
        ],
        '#plan_entity' => $local_plan,
        '#current_user_subscribes_to_any_plan' => $this->subscriptionManager->userHasLocalSubscriptionToAnyPlan($user, 'stripe_connector'),
        '#current_user_subscribes_to_this_plan' => $this->subscriptionManager->userIsSubscribedToPlan($user, $local_plan),
        '#attached' => [
          'library' => [
            'stripe_subscription/checkout',
            'stripe_subscription/stripe.stripejs',
            'core/drupal.dialog.ajax',
          ],
        ],
      ];

      $build['#plans'][$local_plan->getPlanId()] = $element;
    }

    return $build;
  }

}
