<?php

namespace Drupal\stripe_subscription\EventSubscriber;

use Br33f\Ga4\MeasurementProtocol\Dto\Common\UserProperties;
use Br33f\Ga4\MeasurementProtocol\Dto\Common\UserProperty;
use Br33f\Ga4\MeasurementProtocol\Dto\Event\PurchaseEvent;
use Br33f\Ga4\MeasurementProtocol\Dto\Request\BaseRequest;
use Br33f\Ga4\MeasurementProtocol\Service;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\gg_facebook\FacebookConversionApiClient;
use Drupal\stripe_api\Event\StripeApiWebhookEvent;
use Drupal\stripe_api\StripeApiService;
use Drupal\subscription_manager\SubscriptionManagerService;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Stripe\Charge;
use Stripe\Event;
use Stripe\Subscription;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Zumba\Amplitude\Amplitude;

/**
 * Class WebHookSubscriber.
 *
 * @package Drupal\stripe_subscription
 */
class WebHookSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * @var \Drupal\subscription_manager\SubscriptionManagerService
   */
  protected $subscriptionManager;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * @var \Drupal\stripe_subscription\Plugin\SubscriptionManagerConnector\StripeConnector
   */
  private $stripeConnector;

  /**
   * @var \Drupal\stripe_api\StripeApiService
   */
  protected StripeApiService $stripeApi;

  /**
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected AccountProxyInterface $currentUser;

  /**
   * WebHookSubscriber constructor.
   *
   * @param \Drupal\subscription_manager\SubscriptionManagerService $subscription_manager
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   * @param \Drupal\stripe_api\StripeApiService $stripe_api
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function __construct(SubscriptionManagerService $subscription_manager, LoggerChannelInterface $logger, MessengerInterface $messenger, StripeApiService $stripe_api, AccountProxyInterface $current_user) {
    $this->subscriptionManager = $subscription_manager;
    $this->logger = $logger;
    $this->messenger = $messenger;
    $connector_manager = $this->subscriptionManager->getConnectorManager();
    $this->stripeConnector = $connector_manager->createInstance('stripe_connector');
    $this->stripeApi = $stripe_api;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events['stripe_api.webhook'][] = ['onIncomingWebhook'];
    return $events;
  }

  /**
   * Process an incoming webhook.
   *
   * @param \Drupal\stripe_api\Event\StripeApiWebhookEvent $event
   *   Logs an incoming webhook of the setting is on.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Throwable
   */
  public function onIncomingWebhook(StripeApiWebhookEvent $event) {
    $type = $event->type;
    $stripe_event = $event->event;
    $data = $stripe_event->data;
    $this->logStripeEvent($event, $stripe_event);

    // React to subscription life cycle events.
    // @see https://stripe.com/docs/subscriptions/lifecycle
    switch ($type) {
      // Occurs whenever a customer with no subscription is signed up for a plan.
      case 'customer.subscription.created':
        /** @var \Stripe\Subscription $remote_subscription */
        $remote_subscription = $data->object;
        $this->logSubscribeEvent($remote_subscription, $stripe_event);
        // We're actually going to rely on the update event to create the local subscription.
        // $this->createOrUpdateLocalSubscription($remote_subscription);
        break;

      case 'customer.subscription.updated':
        /** @var \Stripe\Subscription $remote_subscription */
        $remote_subscription = $data->object;
        $this->createOrUpdateLocalSubscription($remote_subscription);
        break;

      // Occurs whenever a customer ends their subscription.
      case 'customer.subscription.deleted':
        $remote_subscription = $data->object;
        $this->deleteLocalSubscription($remote_subscription);

        break;

      // Occurs three days before the trial period of a subscription is scheduled to end.
      case 'customer.subscription.trial_will_end':
        break;

      case 'charge.succeeded':
        /** @var \Stripe\Charge $charge */
        $charge = $data->object;

        $customer = $this->stripeApi->getStripeClient()->customers->retrieve($charge->customer);
        if (!$customer) {
          $this->logger->error('Failed to log telemetry events, Customer not found for Stripe customer @customer_id', [
            '@customer_id' => $charge->customer,
          ]);
          return;
        }
        if (!$customer->email) {
          $this->logger->error('Failed to log telemetry events, Customer email not found for Stripe customer @customer_id', [
            '@customer_id' => $charge->customer,
          ]);
          return;
        }
    }

  }

  /**
   * @param \Stripe\Subscription $remote_subscription
   *   The remote subscription.
   *
   * @throws \Throwable
   */
  protected function createLocalSubscription(Subscription $remote_subscription): void {
    try {
      $local_subscription = $this->stripeConnector->createLocalSubscription($remote_subscription);
      $this->messenger->addMessage($this->t('You have successfully subscribed to the @plan_name plan.',
        ['@plan_name' => $remote_subscription->plan->name]), 'status');
      $this->logger->debug('Created local subscription #@subscription_id with remote ID @remote_id',
        ['@subscription_id' => $local_subscription->id(), '@remote_id' => $remote_subscription->id]
      );
    }
    catch (\Throwable $e) {
      $this->logger->error('Failed to create local subscription for remote subscription @remote_id: @exception',
        ['@exception' => $e->getMessage() . $e->getTraceAsString(), '@remote_id' => $remote_subscription->id]
      );
      throw $e;
    }
  }

  /**
   * @param \Stripe\Subscription $remote_subscription
   *
   * @throws \Throwable
   */
  protected function deleteLocalSubscription(Subscription $remote_subscription): void {
    try {
      $this->stripeConnector->syncRemoteSubscriptionToLocal($remote_subscription);
      $local_subscription = $this->subscriptionManager->loadLocalSubscription(['subscription_id' => $remote_subscription->id]);
      $local_subscription->delete();
    }
    catch (\Throwable $e) {
      $this->logger->error('Failed to delete local subscription @remote_id: @exception',
        ['@exception' => $e->getMessage() . $e->getTraceAsString(), '@remote_id' => $remote_subscription->id]
      );
      throw $e;
    }
  }

  /**
   * @param \Drupal\stripe_api\Event\StripeApiWebhookEvent $event
   * @param \Stripe\Event $stripe_event
   *
   * @throws \JsonException
   */
  protected function logStripeEvent(StripeApiWebhookEvent $event, Event $stripe_event): void {
    if (\Drupal::config('stripe_api.settings')->get('log_webhooks')) {
      $this->logger->info("Event Subscriber reacting to @type event:\n @event",
        ['@type' => $event->type, '@event' => json_encode($stripe_event, JSON_THROW_ON_ERROR | JSON_PRETTY_PRINT)]);
    }
  }

  /**
   * @param $remote_subscription
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Throwable
   */
  protected function createOrUpdateLocalSubscription(Subscription $remote_subscription): void {
    $local_subscription = $this->subscriptionManager->loadLocalSubscription(['subscription_id' => $remote_subscription->id]);
    if (!$local_subscription) {
      $this->createLocalSubscription($remote_subscription);
    }
    else {
      $this->stripeConnector->syncRemoteSubscriptionToLocal($remote_subscription);
    }
  }

}
