<?php

namespace Drupal\stripe_subscription\Plugin\Block;

use Drupal\Core\Block\BlockBase;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\stripe_api\StripeApiService;
use Drupal\stripe_subscription\Controller\UserSubscriptionsController;
use Drupal\subscription_manager\SubscriptionManagerService;
use Drupal\user\Entity\User;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a 'SubscribeBlock' block.
 *
 * @Block(
 *  id = "stripe_pricing_table_block",
 *  admin_label = @Translation("Stripe pricing table")
 * )
 *
 * @see https://docs.stripe.com/payments/checkout/pricing-table
 */
class StripePricingTableBlock extends BlockBase implements ContainerFactoryPluginInterface {

  private UserSubscriptionsController $userSubscriptionsController;

  private ImmutableConfig $config;

  private StripeApiService $stripeApi;

  private AccountProxyInterface $currentUser;

  /**
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\subscription_manager\SubscriptionManagerService $subscription_manager
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   * @param \Drupal\stripe_api\StripeApiService $stripe_api
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    SubscriptionManagerService $subscription_manager,
    LoggerChannelInterface $logger,
    AccountProxyInterface $current_user,
    StripeApiService $stripe_api,
    ConfigFactoryInterface $config_factory,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->config = $config_factory->get('stripe_api.settings');
    $this->stripeApi = $stripe_api;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('subscription_manager'),
      $container->get('logger.channel.stripe_subscription'),
      $container->get('current_user'),
      $container->get('stripe_api.stripe_api'),
      $container->get('config.factory'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'price_table_id' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state): array {
    $form['price_table_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Price table ID'),
      '#description' => $this->t('The ID of the price table to display. You can find this in the embed code for the price table.'),
      '#default_value' => $this->configuration['price_table_id'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $this->configuration['price_table_id'] = $values['price_table_id'];
  }

  /**
   * {@inheritdoc}
   */
  public function build($pricing_table_id = NULL) {
    return [
      '#theme' => 'stripe_pricing_table',
      '#pricing_table_id' => $pricing_table_id ?? $this->configuration['price_table_id'],
      '#publishable_key' => $this->stripeApi->getPubKey(),
      '#client_reference_id' => User::load($this->currentUser->id())->uuid(),
      '#customer_email' => $this->currentUser->getEmail(),
      '#attached' => [
        'library' => [
          'stripe_subscription/pricing_table',
        ],
      ],
    ];
  }

}
