<?php

namespace Drupal\stripe_subscription\Plugin\SubscriptionManagerConnector;

use Drupal\Core\Link;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\subscription_manager\Entity\SubscriptionEntity;
use Drupal\subscription_manager\SubscriptionManagerConnectorInterface;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Stripe\Collection;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Class StripeConnector.
 *
 * @SubscriptionManagerConnector(
 *   id = "stripe_connector",
 *   label = @Translation("Stripe Connector"),
 *   description= @Translation("Stripe Connector for Subscription Manager"),
 * )
 *
 * @package Drupal\stripe_subscription
 */
class StripeConnector implements SubscriptionManagerConnectorInterface {

  use StringTranslationTrait;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  private $entityTypeManager;

  /**
   * @var \Drupal\stripe_api\StripeApiService
   */
  protected $stripeApi;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  private $messenger;

  /**
   * @var \Drupal\subscription_manager\SubscriptionManagerService
   */
  private $subscriptionManager;

  /**
   * StripeConnector constructor.
   */
  public function __construct() {
    $this->stripeApi = \Drupal::service('stripe_api.stripe_api');
    $this->entityTypeManager = \Drupal::service('entity_type.manager');
    $this->logger = \Drupal::logger('stripe_subscription');
    $this->messenger = \Drupal::messenger();
    $this->subscriptionManager = \Drupal::service('subscription_manager');
  }

  /**
   * @param \Stripe\Subscription $remote_subscription
   *
   * @return \Drupal\subscription_manager\Entity\SubscriptionEntity
   *   The local subscription entity.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function createLocalSubscription($remote_subscription): SubscriptionEntity {
    $customer = $this->stripeApi->getStripeClient()->customers->retrieve($remote_subscription->customer);
    if (!$customer) {
      throw new \RuntimeException("There is no Stripe customer with id " . $remote_subscription->customer);
    }
    if ($this->stripeApi->getMode() === 'test' &&
      getenv('APP_ENV') === 'local'
      && !$customer->email) {
      $test_email = 'fake-stripe-user@stripe.com';
      if (!user_load_by_mail($test_email)) {
        $this->createUser($test_email);
      }
      $customer->email = $test_email;
    }
    elseif (!$customer->email) {
      throw new \RuntimeException("Stripe customer with id " . $remote_subscription->customer . " has no email address.");
    }

    $user_entity = user_load_by_mail($customer->email);
    if (!$user_entity) {
      throw new \RuntimeException("There is no local user with Stripe with email address " . $customer->email . ". Tried to create local subscription for Stripe subscription ID " . $remote_subscription->id . " for customer with ID " . $remote_subscription->customer);
    }

    $values = [
      'user_id' => $user_entity->id(),
      'plan_id' => $remote_subscription->plan->id,
      'subscription_id' => $remote_subscription->id,
      'customer_id' => $remote_subscription->customer,
      'status' => self::mapStripeSubscriptionStatusToBoolean($remote_subscription->status),
      'connector_plugin_id' => 'stripe_connector',
    ];
    $subscription = $this->entityTypeManager->getStorage('subscription')->create($values);
    $subscription->save();
    $this->logger->info('Created subscription entity #@subscription_id : @sub',
      ['@subscription_id' => $subscription->id(), '@sub' => json_encode($subscription)]
    );

    return $subscription;
  }

  /**
   * Loads a user's remote subscription.
   *
   * @param \Drupal\user\UserInterface|\Drupal\Core\Entity\EntityInterface $user
   *   The user.
   *
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loadRemoteCustomerByUser($user) {
    $customers = $this->stripeApi->getStripeClient()->customers->all(['email' => $user->getEmail()]);

    if (!count($customers->data)) {
      return FALSE;
    }

    return $customers->first();
  }

  /**
   * Loads a user's remote subscription.
   *
   * @param \Drupal\user\UserInterface|\Drupal\Core\Entity\EntityInterface $user
   *   The user.
   *
   * @return bool|\Stripe\Collection
   *   A collection of subscriptions.
   *
   * @throws \Stripe\Error\Api
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loadRemoteSubscriptionsByUser($user) {
    $customer = $this->loadRemoteCustomerByUser($user);
    if (!$customer) {
      return FALSE;
    }
    return $this->loadRemoteSubscriptionMultiple(['customer' => $customer->id]);
  }

  /**
   * Load multiple remote subscriptions.
   *
   * @param array $args
   *   Arguments by which to filter the subscriptions.
   *
   * @return bool|\Stripe\Collection
   *   A collection of subscriptions.
   *
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loadRemoteSubscriptionMultiple(array $args = []): Collection|bool {
    $subscriptions = $this->stripeApi->getStripeClient()->subscriptions->all($args);

    if (!count($subscriptions->data)) {
      return FALSE;
    }

    return $subscriptions;
  }

  /**
   * @param string $plan_id
   *   The plan ID.
   *
   * @return mixed|null
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loadRemotePlanById($plan_id) {
    return $this->stripeApi->getStripeClient()->plans->retrieve($plan_id)->data;
  }

  /**
   * Load multiple remote plans.
   *
   * @param array $args
   *   An array of arguments by which to filter the remote plans.
   *
   * @return \Stripe\Plan[]
   *   An array of remote plans.
   *
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function loadRemotePlanMultiple($args = []): array {
    $plans = $this->stripeApi->getStripeClient()->plans->all($args);

    // @todo handle no results case.
    // Re-key array.
    $keyed_plans = [];
    foreach ($plans->data as $plan) {
      $keyed_plans[$plan->id] = $plan;
    }

    return $keyed_plans;
  }

  /**
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Stripe\Exception\ApiErrorException
   */
  public function syncPlans(): void {
    // @todo Handle pagination here.
    $remote_plans = $this->loadRemotePlanMultiple(['active' => TRUE, 'limit' => 100]);
    $local_plans = $this->entityTypeManager->getStorage('subscription_plan')->loadByProperties(['connector_plugin_id' => 'stripe_connector']);

    /** @var \Drupal\Core\Entity\EntityInterface[] $local_plans_keyed */
    $local_plans_keyed = [];
    foreach ($local_plans as $local_plan) {
      $local_plans_keyed[$local_plan->plan_id->value] = $local_plan;
    }

    $plans_to_disable = array_diff(array_keys($local_plans_keyed), array_keys($remote_plans));
    $plans_to_create = array_diff(array_keys($remote_plans), array_keys($local_plans_keyed));
    $plans_to_update = array_intersect(array_keys($remote_plans), array_keys($local_plans_keyed));

    $this->logger->info('Synchronizing Stripe plans...');

    // Create new plans.
    foreach ($plans_to_create as $plan_id) {
      $data = [
        'plan_id' => $plan_id,
        'name' => $remote_plans[$plan_id]->nickname ?? $remote_plans[$plan_id]->name,
        'data' => $remote_plans[$plan_id]->toArray(),
        'connector_plugin_id' => 'stripe_connector',
      ];
      if ($this->stripeApi->getMode() === 'live') {
        $data['status'] = $remote_plans[$plan_id]->livemode;
      }
      else {
        $data['status'] = !$remote_plans[$plan_id]->livemode;
      }
      $this->entityTypeManager->getStorage('subscription_plan')->create($data)->save();
      $this->logger->info('Created @plan_id plan.', ['@plan_id' => $plan_id]);
    }
    // Disable invalid plans.
    if ($plans_to_disable) {
      foreach ($plans_to_disable as $plan_id) {
        $plan = $local_plans_keyed[$plan_id];
        $plan->set('status', FALSE);
        $plan->save();
        $this->logger->info('Disabled @plan_id plan.', ['@plan_id' => $plan_id]);
      }
    }

    // Update existing plans.
    foreach ($plans_to_update as $plan_id) {
      /** @var \Drupal\Core\Entity\EntityInterface $plan */
      $local_plan = $local_plans_keyed[$plan_id];
      $remote_plan = $remote_plans[$plan_id];
      $local_plan->set('name', $remote_plan->nickname ?? $remote_plan->name);
      $local_plan->set('connector_plugin_id', 'stripe_connector');

      if ($this->stripeApi->getMode() === 'live') {
        $local_plan->set('status', $remote_plan->livemode);
      }
      else {
        $local_plan->set('status', !$remote_plan->livemode);
      }

      $data = $remote_plan->toArray();
      $local_plan->set('data', $data);
      $local_plan->save();
      $this->logger->info('Updated @plan_id plan.', ['@plan_id' => $plan_id]);
    }

    $this->messenger->addMessage($this->t('Plans were synchronized. Visit %link to see synchronized plans. If any plans were newly created, you must assign roles to them.',
      ['%link' => Link::fromTextAndUrl('Subscription plans', Url::fromRoute('entity.subscription_plan.collection'))->toString()]),
       'status');
  }

  /**
   * @param \Stripe\Subscription $remote_subscription
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function syncRemoteSubscriptionToLocal($remote_subscription): void {
    $local_subscription = $this->subscriptionManager->loadLocalSubscription(['subscription_id' => $remote_subscription->id]);
    if (!$local_subscription) {
      throw new \RuntimeException("Could not find matching local subscription for Remote ID $remote_subscription->id.");
    }

    $local_subscription->set('subscription_id', $remote_subscription->id);
    $local_subscription->set('plan_id', $remote_subscription->plan->id);
    $local_subscription->set('customer_id', $remote_subscription->customer);
    $local_subscription->set('status', self::mapStripeSubscriptionStatusToBoolean($remote_subscription->status));
    $local_subscription->set('connector_plugin_id', 'stripe_connector');
    $local_subscription->save();

    $this->logger->info('Updated subscription entity #@subscription_id: @sub', [
      '@subscription_id' => $local_subscription->id(),
      '@sub' => json_encode($local_subscription),
    ]);
  }

  /**
   * @param string $stripe_status
   *   The Stripe subscription status.
   *
   * @return int
   *   The mapped status.
   */
  public static function mapStripeSubscriptionStatusToBoolean(string $stripe_status): int {
    $map = [
      'trialing' => 1,
      'active' => 1,
      'past_due' => 0,
      'canceled' => 0,
      'unpaid' => 0,
      'incomplete' => 0,
    ];

    return $map[$stripe_status];
  }

  /**
   * @param $user
   *
   * @return \Drupal\Core\Routing\TrustedRedirectResponse
   *   The redirect response.
   */
  public function redirectToPortal($user): TrustedRedirectResponse {
    $user_entity = User::load($user);
    $customer_id = $this->subscriptionManager->getLocalUserCustomerIdFromSubscription($user_entity);
    $return_url = Url::fromRoute('<front>', [], ['absolute' => TRUE]);
    // This was not fun.
    // @see https://www.drupal.org/node/2630808
    // @see https://drupal.stackexchange.com/questions/225956/cache-controller-with-json-response
    // @see https://www.lullabot.com/articles/early-rendering-a-lesson-in-debugging-drupal-8
    $return_url_string = $return_url->toString(TRUE)->getGeneratedUrl();
    $session = $this->stripeApi->getStripeClient()->billingPortal->sessions->create([
      'customer' => $customer_id,
      'return_url' => $return_url_string,
    ]);

    return new TrustedRedirectResponse($session->url);
  }

  /**
   * @param $user
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   The redirect response.
   */
  public function redirectToSubscribe($user): RedirectResponse {
    $route_name = 'stripe_subscription.subscribe';
    $route_parameters = [];
    $options = ['absolute' => TRUE];
    $status = 302;
    return new RedirectResponse(Url::fromRoute($route_name, $route_parameters, $options)->toString(), $status);
  }

  /**
   * @param \Drupal\user\UserInterface $user
   *
   * @return bool
   */
  public function shouldRedirectCurrentUserOnLogin(UserInterface $user):bool {
    return TRUE;
  }

  /**
   * @param string $email
   *   The email address.
   *
   * @return \Drupal\user\Entity\User
   *   The user entity.
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function createUser(string $email): User {
    // @see email_registration_user_insert().
    $new_name = preg_replace('/@.*$/', '', $email);
    $new_name = email_registration_cleanup_username($new_name);
    $new_name = email_registration_unique_username($new_name, 0);

    $fields = [
      'name' => $new_name,
      'mail' => $email,
      'init' => $email,
      'pass' => \Drupal::service('password_generator')->generate(32),
      'status' => 1,
    ];

    /** @var \Drupal\user\Entity\User $new_user */
    $entity_type_manager = \Drupal::service('entity_type.manager');
    $new_user = $entity_type_manager
      ->getStorage('user')
      ->create($fields);
    $new_user->save();

    /** @var \Drupal\user\UserDataInterface $userData */
    $user_data = \Drupal::service('user.data');
    // This is used by stripe_subscription_form_user_form_alter().
    $user_data->set('recharge_subscription', $new_user->id(), 'has_random_password', TRUE);

    return $new_user;
  }

}
